#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import argparse
import asyncio
import base64
import functools
import json
import os
import shutil
import sys
import tempfile
import time
import zipfile
from datetime import datetime, timezone
from typing import Awaitable, List, Optional, Set, Union

from aiologger import Logger  # type: ignore
from aiologger.levels import LogLevel  # type: ignore

from aiorun import run  # type: ignore

from dataclasses_jsonschema import ValidationError

import websockets
from websockets.server import WebSocketServerProtocol as WsClient

import arcor2
from arcor2.data import common, compile_json_schemas, events, rpc
from arcor2.data.execution import PackageSummary
from arcor2.data.helpers import EVENT_MAPPING
from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import RPC_DICT_TYPE, aiologger_formatter, read_package_meta, server, write_package_meta
from arcor2.settings import PROJECT_PATH
from arcor2.source.utils import make_executable

PORT = 6790

logger = Logger.with_default_handlers(name='Execution', formatter=aiologger_formatter())

PROCESS: Union[asyncio.subprocess.Process, None] = None
PACKAGE_STATE_EVENT = events.PackageStateEvent()
RUNNING_PACKAGE_ID: Optional[str] = None
PACKAGE_INFO_EVENT: Optional[events.PackageInfoEvent] = None  # in case of man. written scripts, this might not be sent
ACTION_EVENT: Optional[events.ActionStateEvent] = None
ACTION_ARGS_EVENT: Optional[events.CurrentActionEvent] = None
TASK = None

CLIENTS: Set = set()

MAIN_SCRIPT_NAME = "script.py"


def process_running() -> bool:

    return PROCESS is not None and PROCESS.returncode is None


async def package_state(event: events.PackageStateEvent):

    global PACKAGE_STATE_EVENT
    PACKAGE_STATE_EVENT = event
    await send_to_clients(event)


async def read_proc_stdout() -> None:

    global PACKAGE_STATE_EVENT
    global ACTION_EVENT
    global ACTION_ARGS_EVENT
    global PACKAGE_INFO_EVENT
    global RUNNING_PACKAGE_ID

    logger.info("Reading script stdout...")

    assert PROCESS is not None
    assert PROCESS.stdout is not None
    assert RUNNING_PACKAGE_ID is not None

    await package_state(
        events.PackageStateEvent(
            data=common.PackageState(common.PackageStateEnum.RUNNING, RUNNING_PACKAGE_ID)
        )
    )

    printed_out: List[str] = []

    while process_running():
        try:
            stdout = await PROCESS.stdout.readuntil()
        except asyncio.exceptions.IncompleteReadError:
            print("break")
            break

        decoded = stdout.decode("utf-8")
        stripped = decoded.strip()

        try:
            data = json.loads(stripped)
        except json.decoder.JSONDecodeError:
            printed_out.append(decoded)
            logger.error(decoded.strip())
            continue

        if not isinstance(data, dict) or "event" not in data:
            logger.error("Strange data from script: {}".format(data))
            continue

        try:
            evt = EVENT_MAPPING[data["event"]].from_dict(data)
        except ValidationError as e:
            logger.error("Invalid event: {}, error: {}".format(data, e))
            continue

        if isinstance(evt, events.PackageStateEvent):
            evt.data.package_id = RUNNING_PACKAGE_ID
            await package_state(evt)
            continue
        elif isinstance(evt, events.ActionStateEvent):
            ACTION_EVENT = evt
        elif isinstance(evt, events.CurrentActionEvent):
            ACTION_ARGS_EVENT = evt
        elif isinstance(evt, events.PackageInfoEvent):
            PACKAGE_INFO_EVENT = evt

        await send_to_clients(evt)

    ACTION_EVENT = None
    ACTION_ARGS_EVENT = None
    PACKAGE_INFO_EVENT = None

    if PROCESS.returncode:

        if printed_out:

            # TODO remember this (until another package is started) and send it to new clients?
            await send_to_clients(
                events.ProjectExceptionEvent(
                    data=events.ProjectExceptionEventData(printed_out[-1].strip())
                )
            )

            with open("traceback-{}.txt".format(time.strftime("%Y%m%d-%H%M%S")), "w") as tb_file:
                tb_file.write("".join(printed_out))

        else:
            logger.warn(
                f"Process ended with non-zero return code ({PROCESS.returncode}), but didn't printed out anything.")

    await package_state(
        events.PackageStateEvent(
            data=common.PackageState(common.PackageStateEnum.STOPPED, RUNNING_PACKAGE_ID)
        )
    )
    logger.info(f"Process finished with returncode {PROCESS.returncode}.")

    RUNNING_PACKAGE_ID = None


async def run_package_cb(req: rpc.execution.RunPackageRequest, ui: WsClient) -> None:

    global PROCESS
    global TASK
    global RUNNING_PACKAGE_ID

    if process_running():
        raise Arcor2Exception("Already running!")

    package_path = os.path.join(PROJECT_PATH, req.args.id)

    try:
        os.chdir(package_path)
    except FileNotFoundError:
        raise Arcor2Exception("Not found.")

    script_path = os.path.join(package_path, MAIN_SCRIPT_NAME)

    try:
        make_executable(script_path)
    except FileNotFoundError:
        raise Arcor2Exception("Not an execution package.")

    logger.info(f"Starting script: {script_path}")
    PROCESS = await asyncio.create_subprocess_exec(script_path, stdin=asyncio.subprocess.PIPE,
                                                   stdout=asyncio.subprocess.PIPE,
                                                   stderr=asyncio.subprocess.STDOUT)
    if PROCESS.returncode is not None:
        raise Arcor2Exception("Failed to start project.")

    meta = read_package_meta(req.args.id)
    meta.executed = datetime.now(tz=timezone.utc)
    write_package_meta(req.args.id, meta)

    RUNNING_PACKAGE_ID = req.args.id

    TASK = asyncio.ensure_future(read_proc_stdout())  # run task in background


async def stop_package_cb(req: rpc.execution.StopPackageRequest, ui: WsClient) -> None:

    global PACKAGE_INFO_EVENT
    global RUNNING_PACKAGE_ID

    if not process_running():
        raise Arcor2Exception("Project not running.")

    assert PROCESS is not None
    assert TASK is not None

    logger.info("Terminating process")
    PROCESS.terminate()
    logger.info("Waiting for process to finish...")
    await asyncio.wait([TASK])
    PACKAGE_INFO_EVENT = None
    RUNNING_PACKAGE_ID = None


async def pause_package_cb(req: rpc.execution.PausePackageRequest, ui: WsClient) -> None:

    if not process_running():
        raise Arcor2Exception("Project not running.")

    assert PROCESS is not None
    assert PROCESS.stdin is not None

    if PACKAGE_STATE_EVENT.data.state != common.PackageStateEnum.RUNNING:
        raise Arcor2Exception("Cannot pause.")

    PROCESS.stdin.write("p\n".encode())
    await PROCESS.stdin.drain()
    return None


async def resume_package_cb(req: rpc.execution.ResumePackageRequest, ui: WsClient) -> None:

    if not process_running():
        raise Arcor2Exception("Project not running.")

    assert PROCESS is not None
    assert PROCESS.stdin is not None

    if PACKAGE_STATE_EVENT.data.state != common.PackageStateEnum.PAUSED:
        raise Arcor2Exception("Cannot resume.")

    PROCESS.stdin.write("r\n".encode())
    await PROCESS.stdin.drain()
    return None


async def package_state_cb(req: rpc.execution.PackageStateRequest, ui: WsClient) ->\
        rpc.execution.PackageStateResponse:

    resp = rpc.execution.PackageStateResponse()
    resp.data.project = PACKAGE_STATE_EVENT.data
    if ACTION_EVENT:
        resp.data.action = ACTION_EVENT.data
    if ACTION_ARGS_EVENT:
        resp.data.action_args = ACTION_ARGS_EVENT.data
    return resp


async def _upload_package_cb(req: rpc.execution.UploadPackageRequest, ui: WsClient) -> None:

    target_path = os.path.join(PROJECT_PATH, req.args.id)

    # TODO do not allow if there are manual changes?

    with tempfile.TemporaryDirectory() as tmpdirname:

        zip_path = os.path.join(tmpdirname, "publish.zip")

        b64_bytes = req.args.data.encode()
        zip_content = base64.b64decode(b64_bytes)

        with open(zip_path, "wb") as zip_file:
            zip_file.write(zip_content)

        try:
            with zipfile.ZipFile(zip_path, 'r') as zip_ref:
                zip_ref.extractall(tmpdirname)
        except zipfile.BadZipFile as e:
            logger.error(e)
            raise Arcor2Exception("Invalid zip file.")

        os.remove(zip_path)

        try:
            shutil.rmtree(target_path)
        except FileNotFoundError:
            pass
        shutil.copytree(tmpdirname, target_path)

    script_path = os.path.join(target_path, MAIN_SCRIPT_NAME)

    try:
        make_executable(script_path)
    except FileNotFoundError:
        raise Arcor2Exception("Package does not contain 'script.py' file.")

    asyncio.ensure_future(
        send_to_clients(
            events.PackageChanged(events.EventType.ADD, data=await get_summary(target_path))
        )
    )
    return None


async def get_summary(path: str) -> PackageSummary:

    if not os.path.isfile(os.path.join(path, MAIN_SCRIPT_NAME)):
        raise Arcor2Exception("Package does not contain main script.")

    package_dir = os.path.basename(path)
    package_meta = read_package_meta(package_dir)

    try:
        with open(os.path.join(path, "data", "project.json")) as project_file:
            project = common.Project.from_json(project_file.read())
    except (ValidationError, IOError) as e:
        logger.error(f"Failed to read/parse project file of {package_dir}: {e}")

        return PackageSummary(package_dir, "N/A", datetime.fromtimestamp(0, tz=timezone.utc), package_meta)

    assert project.modified
    return PackageSummary(package_dir, project.id, project.modified, package_meta)


async def list_packages_cb(req: rpc.execution.ListPackagesRequest, ui: WsClient) ->\
        rpc.execution.ListPackagesResponse:

    resp = rpc.execution.ListPackagesResponse()

    subfolders = [f.path for f in os.scandir(PROJECT_PATH) if f.is_dir()]

    for folder_path in subfolders:

        try:
            resp.data.append(await get_summary(folder_path))
        except Arcor2Exception:
            pass

    return resp


async def delete_package_cb(req: rpc.execution.DeletePackageRequest, ui: WsClient) -> None:

    if RUNNING_PACKAGE_ID and RUNNING_PACKAGE_ID == req.args.id:
        raise Arcor2Exception("Package is being executed.")

    target_path = os.path.join(PROJECT_PATH, req.args.id)
    package_summary = await get_summary(target_path)

    try:
        shutil.rmtree(target_path)
    except FileNotFoundError:
        raise Arcor2Exception("Not found.")

    asyncio.ensure_future(send_to_clients(events.PackageChanged(events.EventType.REMOVE, data=package_summary)))
    return None


async def rename_package_cb(req: rpc.execution.RenamePackageRequest, ui: WsClient) -> None:

    target_path = os.path.join(PROJECT_PATH, req.args.package_id, "package.json")

    pm = read_package_meta(req.args.package_id)
    pm.name = req.args.new_name

    with open(target_path, "w") as pkg_file:
        pkg_file.write(pm.to_json())

    asyncio.ensure_future(
        send_to_clients(
            events.PackageChanged(events.EventType.UPDATE,
                                  data=await get_summary(os.path.join(PROJECT_PATH, req.args.package_id)))
        )
    )


async def _version_cb(req: rpc.common.VersionRequest, ui: WsClient) -> rpc.common.VersionResponse:
    return rpc.common.VersionResponse(data=rpc.common.VersionData(arcor2.api_version()))


async def send_to_clients(event: events.Event) -> None:

    if CLIENTS:
        data = event.to_json()
        await asyncio.wait([client.send(data) for client in CLIENTS])


async def register(websocket: WsClient) -> None:

    logger.info("Registering new client")
    CLIENTS.add(websocket)

    tasks: List[Awaitable] = [websocket.send(PACKAGE_STATE_EVENT.to_json())]

    if PACKAGE_INFO_EVENT:
        tasks.append(websocket.send(PACKAGE_INFO_EVENT.to_json()))

    await asyncio.gather(*tasks)


async def unregister(websocket: WsClient) -> None:
    logger.info("Unregistering client")
    CLIENTS.remove(websocket)

RPC_DICT: RPC_DICT_TYPE = {
    rpc.execution.RunPackageRequest: run_package_cb,
    rpc.execution.StopPackageRequest: stop_package_cb,
    rpc.execution.PausePackageRequest: pause_package_cb,
    rpc.execution.ResumePackageRequest: resume_package_cb,
    rpc.execution.PackageStateRequest: package_state_cb,
    rpc.execution.UploadPackageRequest: _upload_package_cb,
    rpc.execution.ListPackagesRequest: list_packages_cb,
    rpc.execution.DeletePackageRequest: delete_package_cb,
    rpc.execution.RenamePackageRequest: rename_package_cb,
    rpc.common.VersionRequest: _version_cb
}


async def aio_main() -> None:

    await websockets.serve(
        functools.partial(server, logger=logger, register=register, unregister=unregister, rpc_dict=RPC_DICT),
        '0.0.0.0', PORT)


def main() -> None:

    assert sys.version_info >= (3, 8)

    parser = argparse.ArgumentParser()

    parser.add_argument("-v", "--verbose", help="Increase output verbosity",
                        action="store_const", const=LogLevel.DEBUG, default=LogLevel.INFO)
    parser.add_argument('--version', action='version', version=arcor2.version(),
                        help="Shows ARCOR2 version and exits.")
    parser.add_argument('--api_version', action='version', version=arcor2.api_version(),
                        help="Shows API version and exits.")
    parser.add_argument("-a", "--asyncio_debug", help="Turn on asyncio debug mode.",
                        action="store_const", const=True, default=False)

    args = parser.parse_args()
    logger.level = args.verbose

    loop = asyncio.get_event_loop()
    loop.set_debug(enabled=args.asyncio_debug)

    compile_json_schemas()

    run(aio_main(), loop=loop, stop_on_unhandled_errors=True)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

from dataclasses import dataclass, field
from typing import List, Set

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data.common import IdValue, StrEnum
from arcor2.data.object_type import ObjectActions, ObjectTypeMeta
from arcor2.data.rpc.common import IdArgs, Request, Response, RobotArg, TypeArgs, wo_suffix


@dataclass
class ActionParamValuesArgs(JsonSchemaMixin):

    id: str = field(metadata=dict(description="Object or service id."))
    param_id: str
    parent_params: List[IdValue] = field(default_factory=list)


@dataclass
class ActionParamValuesRequest(Request):

    args: ActionParamValuesArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ActionParamValuesResponse(Response):

    data: Set[str] = field(default_factory=set)  # TODO what about other (possible) types?
    response: str = field(default=ActionParamValuesRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class GetObjectTypesRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class GetObjectTypesResponse(Response):

    data: List[ObjectTypeMeta] = field(default_factory=list)
    response: str = field(default=GetObjectTypesRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class DeleteObjectTypeRequest(Request):

    args: IdArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class DeleteObjectTypeResponse(Response):

    response: str = field(default=DeleteObjectTypeRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class GetActionsRequest(Request):

    args: TypeArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class GetActionsResponse(Response):

    data: ObjectActions = field(default_factory=list)
    response: str = field(default=GetActionsRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class NewObjectTypeRequest(Request):

    args: ObjectTypeMeta
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class NewObjectTypeResponse(Response):

    response: str = field(default=NewObjectTypeRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class FocusObjectStartRequestArgs(JsonSchemaMixin):

    object_id: str
    robot: RobotArg


@dataclass
class FocusObjectStartRequest(Request):

    args: FocusObjectStartRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class FocusObjectStartResponse(Response):

    response: str = field(default=FocusObjectStartRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class FocusObjectRequestArgs(JsonSchemaMixin):

    object_id: str
    point_idx: int


@dataclass
class FocusObjectRequest(Request):

    args: FocusObjectRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class FocusObjectResponseData(JsonSchemaMixin):

    finished_indexes: List[int] = field(default_factory=list)


@dataclass
class FocusObjectResponse(Response):

    data: FocusObjectResponseData = field(default_factory=FocusObjectResponseData)
    response: str = field(default=FocusObjectRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class FocusObjectDoneRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class FocusObjectDoneResponse(Response):

    data: FocusObjectResponseData = field(default_factory=FocusObjectResponseData)
    response: str = field(default=FocusObjectDoneRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------

class PivotEnum(StrEnum):
    TOP: str = "top"
    MIDDLE: str = "middle"
    BOTTOM: str = "bottom"


@dataclass
class UpdateObjectPoseUsingRobotArgs(IdArgs):

    robot: RobotArg
    pivot: PivotEnum = PivotEnum.MIDDLE


@dataclass
class UpdateObjectPoseUsingRobotRequest(Request):

    args: UpdateObjectPoseUsingRobotArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateObjectPoseUsingRobotResponse(Response):

    response: str = field(default=UpdateObjectPoseUsingRobotRequest.request, init=False)



# -*- coding: utf-8 -*-

from dataclasses import dataclass, field
from typing import List, Optional, Set

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data import common, robot
from arcor2.data.rpc.common import Request, Response, wo_suffix


@dataclass
class GetRobotMetaRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class GetRobotMetaResponse(Response):

    data: List[robot.RobotMeta] = field(default_factory=list)
    response: str = field(default=GetRobotMetaRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class GetRobotJointsArgs(JsonSchemaMixin):

    robot_id: str


@dataclass
class GetRobotJointsRequest(Request):

    args: GetRobotJointsArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class GetRobotJointsResponse(Response):

    data: List[common.Joint] = field(default_factory=list)
    response: str = field(default=GetRobotJointsRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class GetEndEffectorPoseArgs(JsonSchemaMixin):

    robot_id: str
    end_effector_id: str


@dataclass
class GetEndEffectorPoseRequest(Request):

    args: GetEndEffectorPoseArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class GetEndEffectorPoseResponse(Response):

    data: common.Pose = field(default_factory=common.Pose)
    response: str = field(default=GetEndEffectorPoseRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class GetEndEffectorsArgs(JsonSchemaMixin):

    robot_id: str


@dataclass
class GetEndEffectorsRequest(Request):

    args: GetEndEffectorsArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class GetEndEffectorsResponse(Response):

    data: Set[str] = field(default_factory=set)
    response: str = field(default=GetEndEffectorsRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class GetGrippersArgs(JsonSchemaMixin):

    robot_id: str


@dataclass
class GetGrippersRequest(Request):

    args: GetGrippersArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class GetGrippersResponse(Response):

    data: Set[str] = field(default_factory=set)
    response: str = field(default=GetGrippersRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class GetSuctionsArgs(JsonSchemaMixin):

    robot_id: str


@dataclass
class GetSuctionsRequest(Request):

    args: GetSuctionsArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class GetSuctionsResponse(Response):

    data: Set[str] = field(default_factory=set)
    response: str = field(default=GetSuctionsRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


class RegisterEnum(common.StrEnum):

    EEF_POSE: str = "eef_pose"
    JOINTS: str = "joints"


@dataclass
class RegisterForRobotEventArgs(JsonSchemaMixin):

    robot_id: str
    what: RegisterEnum
    send: bool


@dataclass
class RegisterForRobotEventRequest(Request):

    args: RegisterForRobotEventArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RegisterForRobotEventResponse(Response):

    data: Set[str] = field(default_factory=set)
    response: str = field(default=RegisterForRobotEventRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class MoveToPoseArgs(JsonSchemaMixin):

    robot_id: str
    end_effector_id: str
    speed: float
    position: Optional[common.Position]
    orientation: Optional[common.Orientation]


@dataclass
class MoveToPoseRequest(Request):

    args: MoveToPoseArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class MoveToPoseResponse(Response):
    response: str = field(default=MoveToPoseRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class StopRobotArgs(JsonSchemaMixin):

    robot_id: str


@dataclass
class StopRobotRequest(Request):

    args: StopRobotArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class StopRobotResponse(Response):
    response: str = field(default=StopRobotRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class MoveToJointsArgs(JsonSchemaMixin):

    robot_id: str
    speed: float
    joints: List[common.Joint]


@dataclass
class MoveToJointsRequest(Request):

    args: MoveToJointsArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class MoveToJointsResponse(Response):
    response: str = field(default=MoveToJointsRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class MoveToActionPointArgs(JsonSchemaMixin):

    robot_id: str
    speed: float
    end_effector_id: Optional[str] = None
    orientation_id: Optional[str] = None
    joints_id: Optional[str] = None


@dataclass
class MoveToActionPointRequest(Request):

    args: MoveToActionPointArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class MoveToActionPointResponse(Response):
    response: str = field(default=MoveToActionPointRequest.request, init=False)



import inspect

from apispec import APISpec  # type: ignore
from apispec.exceptions import DuplicateComponentNameError  # type: ignore

from apispec_webframeworks.flask import FlaskPlugin  # type: ignore

from dataclasses_jsonschema import JsonSchemaMixin
from dataclasses_jsonschema.apispec import DataclassesPlugin

import arcor2
import arcor2.data.common
import arcor2.data.events
import arcor2.data.object_type
from arcor2.data import rpc


def generate_swagger() -> str:

    # Create an APISpec
    spec = APISpec(
        title="ARCOR2 Data Models",
        version=arcor2.api_version(),
        openapi_version="3.0.2",
        plugins=[FlaskPlugin(), DataclassesPlugin()],
    )

    # TODO avoid explicit naming of all sub-modules in rpc module
    for module in (arcor2.data.common, arcor2.data.object_type, rpc.common, rpc.execution, rpc.objects,
                   rpc.robot, rpc.scene, rpc.project, rpc.storage, arcor2.data.events):
        for _, obj in inspect.getmembers(module):

            if not inspect.isclass(obj) or not issubclass(obj, JsonSchemaMixin) or obj == JsonSchemaMixin:
                continue

            try:
                spec.components.schema(obj.__name__, schema=obj)
            except DuplicateComponentNameError:
                continue

    return spec.to_yaml()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import argparse
import asyncio
import functools
import inspect
import json
import os
import shutil
import sys
import uuid
from typing import get_type_hints

from aiologger.levels import LogLevel  # type: ignore

from aiorun import run  # type: ignore

from dataclasses_jsonschema import ValidationError

import websockets
from websockets.server import WebSocketServerProtocol as WsClient

import arcor2
import arcor2.helpers as hlp
from arcor2 import action as action_mod
from arcor2.clients import aio_persistent_storage as storage
from arcor2.data import common, compile_json_schemas, events
from arcor2.data import rpc
from arcor2.data.helpers import EVENT_MAPPING, RPC_MAPPING
from arcor2.exceptions import Arcor2Exception
from arcor2.nodes.execution import RPC_DICT as EXE_RPC_DICT
from arcor2.parameter_plugins import PARAM_PLUGINS
from arcor2.server import events as server_events, execution as exe, globals as glob, notifications as notif, \
    objects_services_actions as osa, rpc as srpc, settings
from arcor2.source.object_types import prepare_object_types_dir

# disables before/after messages, etc.
action_mod.HANDLE_ACTIONS = False


async def handle_manager_incoming_messages(manager_client) -> None:

    try:

        async for message in manager_client:

            msg = json.loads(message)

            if "event" in msg:

                if glob.INTERFACES:
                    await asyncio.gather(*[hlp.send_json_to_client(intf, message) for intf in glob.INTERFACES])

                try:
                    evt = EVENT_MAPPING[msg["event"]].from_dict(msg)
                except ValidationError as e:
                    glob.logger.error("Invalid event: {}, error: {}".format(msg, e))
                    continue

                if isinstance(evt, events.PackageInfoEvent):
                    glob.PACKAGE_INFO = evt.data
                elif isinstance(evt, events.PackageStateEvent):
                    glob.PACKAGE_STATE = evt.data

                    if evt.data.state == common.PackageStateEnum.STOPPED:

                        if not glob.TEMPORARY_PACKAGE:
                            # after (ordinary) package is finished, show list of packages
                            glob.MAIN_SCREEN = \
                                events.ShowMainScreenData(events.ShowMainScreenData.WhatEnum.PackagesList)
                            await notif.broadcast_event(events.ShowMainScreenEvent(
                                data=events.ShowMainScreenData(events.ShowMainScreenData.WhatEnum.PackagesList,
                                                               evt.data.package_id)))

                        # temporary package is handled elsewhere
                        server_events.package_stopped.set()
                        server_events.package_started.clear()

                        glob.CURRENT_ACTION = None
                        glob.ACTION_STATE = None
                        glob.PACKAGE_INFO = None

                    else:
                        server_events.package_stopped.clear()
                        server_events.package_started.set()

                elif isinstance(evt, events.ActionStateEvent):
                    glob.ACTION_STATE = evt.data
                elif isinstance(evt, events.CurrentActionEvent):
                    glob.CURRENT_ACTION = evt.data

            elif "response" in msg:

                # TODO handle potential errors
                _, resp_cls = RPC_MAPPING[msg["response"]]
                resp = resp_cls.from_dict(msg)
                exe.MANAGER_RPC_RESPONSES[resp.id].put_nowait(resp)

    except websockets.exceptions.ConnectionClosed:
        glob.logger.error("Connection to manager closed.")


async def _initialize_server() -> None:

    exe_version = await exe.manager_request(rpc.common.VersionRequest(uuid.uuid4().int))
    assert isinstance(exe_version, rpc.common.VersionResponse)

    """
    Following check is especially useful when running server/execution in Docker containers.
    Then it might easily happen that one tries to use different versions together.
    """
    try:
        hlp.check_compatibility(arcor2.api_version(), exe_version.data.version)
    except Arcor2Exception as e:
        raise Arcor2Exception("ARServer/Execution API_VERSION mismatch.") from e

    while True:  # wait until Project service becomes available
        try:
            await storage.get_projects()
            break
        except storage.PersistentStorageException as e:
            print(e.message)
            await asyncio.sleep(1)

    await osa.get_object_types()

    bound_handler = functools.partial(hlp.server, logger=glob.logger, register=register, unregister=unregister,
                                      rpc_dict=RPC_DICT, event_dict=EVENT_DICT, verbose=glob.VERBOSE)

    glob.logger.info("Server initialized.")
    await asyncio.wait([websockets.serve(bound_handler, '0.0.0.0', glob.PORT)])


async def list_meshes_cb(req: rpc.storage.ListMeshesRequest, ui: WsClient) -> rpc.storage.ListMeshesResponse:
    return rpc.storage.ListMeshesResponse(data=await storage.get_meshes())


async def register(websocket: WsClient) -> None:

    glob.logger.info("Registering new ui")
    glob.INTERFACES.add(websocket)

    if glob.PROJECT:
        assert glob.SCENE
        await notif.event(websocket,
                          events.OpenProject(data=events.OpenProjectData(glob.SCENE.scene, glob.PROJECT.project)))
    elif glob.SCENE:
        await notif.event(websocket, events.OpenScene(data=events.OpenSceneData(glob.SCENE.scene)))
    elif glob.PACKAGE_INFO:

        # this can't be done in parallel - ui expects this order of events
        await websocket.send(events.PackageStateEvent(data=glob.PACKAGE_STATE).to_json())
        await websocket.send(events.PackageInfoEvent(data=glob.PACKAGE_INFO).to_json())

        if glob.ACTION_STATE:
            await websocket.send(events.ActionStateEvent(data=glob.ACTION_STATE).to_json())
        if glob.CURRENT_ACTION:
            await websocket.send(events.CurrentActionEvent(data=glob.CURRENT_ACTION).to_json())
    else:
        await notif.event(websocket, events.ShowMainScreenEvent(data=glob.MAIN_SCREEN))


async def unregister(websocket: WsClient) -> None:
    glob.logger.info("Unregistering ui")  # TODO print out some identifier
    glob.INTERFACES.remove(websocket)

    for registered_uis in glob.ROBOT_JOINTS_REGISTERED_UIS.values():
        if websocket in registered_uis:
            registered_uis.remove(websocket)
    for registered_uis in glob.ROBOT_EEF_REGISTERED_UIS.values():
        if websocket in registered_uis:
            registered_uis.remove(websocket)


async def system_info_cb(req: rpc.common.SystemInfoRequest, ui: WsClient) -> rpc.common.SystemInfoResponse:

    resp = rpc.common.SystemInfoResponse()
    resp.data.version = arcor2.version()
    resp.data.api_version = arcor2.api_version()
    resp.data.supported_parameter_types = set(PARAM_PLUGINS.keys())
    resp.data.supported_rpc_requests = {key.request for key in RPC_DICT.keys()}
    return resp


RPC_DICT: hlp.RPC_DICT_TYPE = {
    rpc.common.SystemInfoRequest: system_info_cb
}

# discovery of RPC callbacks
for _, rpc_module in inspect.getmembers(srpc, inspect.ismodule):
    for rpc_cb_name, rpc_cb in inspect.getmembers(rpc_module):

        if not rpc_cb_name.endswith("_cb"):
            continue

        hints = get_type_hints(rpc_cb)

        try:
            ttype = hints["req"]
        except KeyError:
            continue

        RPC_DICT[ttype] = rpc_cb

# add Project Manager RPC API
for k, v in EXE_RPC_DICT.items():

    if v.__name__.startswith("_"):
        continue

    RPC_DICT[k] = exe.manager_request


# events from clients
EVENT_DICT: hlp.EVENT_DICT_TYPE = {}


async def aio_main() -> None:

    await asyncio.gather(
        exe.project_manager_client(handle_manager_incoming_messages),
        _initialize_server()
    )


def main() -> None:

    assert sys.version_info >= (3, 8)

    parser = argparse.ArgumentParser()

    parser.add_argument("-v", "--verbose", help="Increase verbosity.",
                        action="store_const", const=True, default=False)
    parser.add_argument("-d", "--debug", help="Set logging level to debug.",
                        action="store_const", const=LogLevel.DEBUG, default=LogLevel.INFO)
    parser.add_argument('--version', action='version', version=arcor2.version(),
                        help="Shows ARCOR2 version and exits.")
    parser.add_argument('--api_version', action='version', version=arcor2.api_version(),
                        help="Shows API version and exits.")
    parser.add_argument("-a", "--asyncio_debug", help="Turn on asyncio debug mode.",
                        action="store_const", const=True, default=False)

    args = parser.parse_args()
    glob.logger.level = args.debug
    glob.VERBOSE = args.verbose

    loop = asyncio.get_event_loop()
    loop.set_debug(enabled=args.asyncio_debug)

    compile_json_schemas()

    if os.path.exists(settings.URDF_PATH):
        shutil.rmtree(settings.URDF_PATH)
    os.makedirs(settings.URDF_PATH)

    prepare_object_types_dir(settings.OBJECT_TYPE_PATH, settings.OBJECT_TYPE_MODULE)

    run(aio_main(), loop=loop, stop_on_unhandled_errors=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import asyncio
from typing import Dict, List

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2.cached import CachedScene
from arcor2.clients import aio_persistent_storage as storage, aio_scene_service as scene_srv
from arcor2.data import events, rpc
from arcor2.data.common import Pose, Position
from arcor2.data.object_type import MeshFocusAction, Model3dType
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types import utils as otu
from arcor2.object_types.abstract import GenericWithPose, Robot
from arcor2.parameter_plugins import TYPE_TO_PLUGIN
from arcor2.server import globals as glob, notifications as notif, objects_services_actions as osa, settings
from arcor2.server.decorators import no_project, scene_needed
from arcor2.server.project import scene_object_pose_updated
from arcor2.server.robot import get_end_effector_pose
from arcor2.server.scene import scenes, set_object_pose
from arcor2.source.object_types import new_object_type
from arcor2.source.utils import tree_to_str

FOCUS_OBJECT: Dict[str, Dict[int, Pose]] = {}  # object_id / idx, pose
FOCUS_OBJECT_ROBOT: Dict[str, rpc.common.RobotArg] = {}  # key: object_id


def clean_up_after_focus(obj_id: str) -> None:

    try:
        del FOCUS_OBJECT[obj_id]
    except KeyError:
        pass

    try:
        del FOCUS_OBJECT_ROBOT[obj_id]
    except KeyError:
        pass


@scene_needed
@no_project
async def focus_object_start_cb(req: rpc.objects.FocusObjectStartRequest, ui: WsClient) -> None:

    obj_id = req.args.object_id

    if obj_id in FOCUS_OBJECT_ROBOT:
        raise Arcor2Exception("Focusing already started.")

    if obj_id not in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Unknown object.")

    inst = await osa.get_robot_instance(req.args.robot.robot_id, req.args.robot.end_effector)

    robot_type = glob.OBJECT_TYPES[inst.__class__.__name__]
    assert robot_type.robot_meta

    obj_type = glob.OBJECT_TYPES[osa.get_obj_type_name(obj_id)].meta

    if not obj_type.has_pose:
        raise Arcor2Exception("Only available for objects with pose.")

    if not obj_type.object_model or obj_type.object_model.type != Model3dType.MESH:
        raise Arcor2Exception("Only available for objects with mesh model.")

    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    if not focus_points:
        raise Arcor2Exception("focusPoints not defined for the mesh.")

    FOCUS_OBJECT_ROBOT[req.args.object_id] = req.args.robot
    FOCUS_OBJECT[obj_id] = {}
    glob.logger.info(f'Start of focusing for {obj_id}.')
    return None


@no_project
async def focus_object_cb(req: rpc.objects.FocusObjectRequest, ui: WsClient) -> rpc.objects.FocusObjectResponse:

    obj_id = req.args.object_id
    pt_idx = req.args.point_idx

    if obj_id not in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Unknown object_id.")

    obj_type = glob.OBJECT_TYPES[osa.get_obj_type_name(obj_id)].meta

    assert obj_type.has_pose
    assert obj_type.object_model
    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    assert focus_points

    if pt_idx < 0 or pt_idx > len(focus_points) - 1:
        raise Arcor2Exception("Index out of range.")

    if obj_id not in FOCUS_OBJECT:
        glob.logger.info(f'Start of focusing for {obj_id}.')
        FOCUS_OBJECT[obj_id] = {}

    robot_id, end_effector = FOCUS_OBJECT_ROBOT[obj_id].as_tuple()

    FOCUS_OBJECT[obj_id][pt_idx] = await get_end_effector_pose(robot_id, end_effector)

    r = rpc.objects.FocusObjectResponse()
    r.data.finished_indexes = list(FOCUS_OBJECT[obj_id].keys())
    return r


@scene_needed
@no_project
async def focus_object_done_cb(req: rpc.objects.FocusObjectDoneRequest, ui: WsClient) -> None:

    obj_id = req.args.id

    if obj_id not in FOCUS_OBJECT:
        raise Arcor2Exception("focusObjectStart/focusObject has to be called first.")

    obj_type = glob.OBJECT_TYPES[osa.get_obj_type_name(obj_id)].meta

    assert obj_type.object_model
    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    assert focus_points

    if len(FOCUS_OBJECT[obj_id]) < len(focus_points):
        raise Arcor2Exception("Not all points were done.")

    assert glob.SCENE

    obj = glob.SCENE.object(obj_id)
    assert obj.pose

    obj_inst = glob.SCENE_OBJECT_INSTANCES[obj_id]
    assert isinstance(obj_inst, GenericWithPose)

    fp: List[Position] = []
    rp: List[Position] = []

    for idx, pose in FOCUS_OBJECT[obj_id].items():

        fp.append(focus_points[idx].position)
        rp.append(pose.position)

    mfa = MeshFocusAction(fp, rp)

    glob.logger.debug(f'Attempt to focus for object {obj_id}, data: {mfa}')

    try:
        new_pose = await scene_srv.focus(mfa)
    except scene_srv.SceneServiceException as e:
        glob.logger.error(f"Focus failed with: {e}, mfa: {mfa}.")
        raise Arcor2Exception("Focusing failed.") from e

    obj.pose = new_pose
    glob.SCENE.update_modified()

    glob.logger.info(f"Done focusing for {obj_id}.")

    clean_up_after_focus(obj_id)

    asyncio.ensure_future(notif.broadcast_event(events.SceneObjectChanged(events.EventType.UPDATE, data=obj)))
    asyncio.ensure_future(scene_object_pose_updated(glob.SCENE.id, obj.id))
    asyncio.ensure_future(set_object_pose(obj_inst, new_pose))

    return None


async def new_object_type_cb(req: rpc.objects.NewObjectTypeRequest, ui: WsClient) -> None:

    meta = req.args

    if meta.type in glob.OBJECT_TYPES:
        raise Arcor2Exception("Object type already exists.")

    if not hlp.is_valid_type(meta.type):
        raise Arcor2Exception("Object type invalid (should be CamelCase).")

    if meta.base not in glob.OBJECT_TYPES:
        raise Arcor2Exception(f"Unknown base object type '{meta.base}', "
                              f"known types are: {', '.join(glob.OBJECT_TYPES.keys())}.")

    base = glob.OBJECT_TYPES[meta.base]

    if base.meta.disabled:
        raise Arcor2Exception("Base object is disabled.")

    assert base.type_def is not None

    if issubclass(base.type_def, Robot):
        raise Arcor2Exception("Can't subclass Robot.")

    meta.has_pose = issubclass(base.type_def, GenericWithPose)

    if not meta.has_pose and meta.object_model:
        raise Arcor2Exception("Object without pose can't have collision model.")

    if req.dry_run:
        return None

    obj = meta.to_object_type()
    ast = new_object_type(glob.OBJECT_TYPES[meta.base].meta, meta)
    obj.source = tree_to_str(ast)

    if meta.object_model and meta.object_model.type != Model3dType.MESH:
        assert meta.type == meta.object_model.model().id
        await storage.put_model(meta.object_model.model())

    # TODO check whether mesh id exists - if so, then use existing mesh, if not, upload a new one
    if meta.object_model and meta.object_model.type == Model3dType.MESH:
        # ...get whole mesh (focus_points) based on mesh id
        assert meta.object_model.mesh
        try:
            meta.object_model.mesh = await storage.get_mesh(meta.object_model.mesh.id)
        except storage.PersistentStorageException as e:
            glob.logger.error(e)
            raise Arcor2Exception(f"Mesh ID {meta.object_model.mesh.id} does not exist.")

    type_def = await hlp.run_in_executor(
        hlp.save_and_import_type_def, obj.source, obj.id, base.type_def, settings.OBJECT_TYPE_PATH,
        settings.OBJECT_TYPE_MODULE)
    assert issubclass(type_def, base.type_def)
    actions = otu.object_actions(TYPE_TO_PLUGIN, type_def, ast)

    await storage.update_object_type(obj)

    glob.OBJECT_TYPES[meta.type] = otu.ObjectTypeData(meta, type_def, actions, ast)
    otu.add_ancestor_actions(meta.type, glob.OBJECT_TYPES)

    asyncio.ensure_future(notif.broadcast_event(events.ChangedObjectTypesEvent(events.EventType.ADD, data=[meta])))
    return None


async def get_object_actions_cb(req: rpc.objects.GetActionsRequest, ui: WsClient) -> rpc.objects.GetActionsResponse:

    try:
        return rpc.objects.GetActionsResponse(data=list(glob.OBJECT_TYPES[req.args.type].actions.values()))
    except KeyError:
        raise Arcor2Exception(f"Unknown object type: '{req.args.type}'.")


async def get_object_types_cb(req: rpc.objects.GetObjectTypesRequest, ui: WsClient) ->\
        rpc.objects.GetObjectTypesResponse:
    return rpc.objects.GetObjectTypesResponse(data=[obj.meta for obj in glob.OBJECT_TYPES.values()])


def check_scene_for_object_type(scene: CachedScene, object_type: str) -> None:

    if list(scene.objects_of_type(object_type)):
        raise Arcor2Exception(f"Object type used in scene '{scene.name}'.")


async def delete_object_type_cb(req: rpc.objects.DeleteObjectTypeRequest, ui: WsClient) -> None:

    try:
        obj_type = glob.OBJECT_TYPES[req.args.id]
    except KeyError:
        raise Arcor2Exception("Unknown object type.")

    if obj_type.meta.built_in:
        raise Arcor2Exception("Can't delete built-in type.")

    for obj in glob.OBJECT_TYPES.values():
        if obj.meta.base == req.args.id:
            raise Arcor2Exception(f"Object type is base of '{obj.meta.type}'.")

    async for scene in scenes():
        check_scene_for_object_type(scene, req.args.id)

    if glob.SCENE:
        check_scene_for_object_type(glob.SCENE, req.args.id)

    if req.dry_run:
        return

    await storage.delete_object_type(req.args.id)

    # do not care so much if delete_model fails
    if obj_type.meta.object_model:
        try:
            await storage.delete_model(obj_type.meta.object_model.model().id)
        except storage.PersistentStorageException as e:
            glob.logger.error(e.message)

    del glob.OBJECT_TYPES[req.args.id]
    asyncio.ensure_future(notif.broadcast_event(
        events.ChangedObjectTypesEvent(events.EventType.REMOVE, data=[obj_type.meta])))



def _get_message(e: Exception) -> str:

    try:
        arg = e.args[0]
        if isinstance(arg, Exception):
            return _get_message(arg)
        return arg
    except IndexError:
        return "No message available."


class Arcor2Exception(Exception):
    """
    All exceptions are derived from this one.
    """

    @property
    def message(self) -> str:
        """
        Gets the last (most user-oriented) message from the exception.
        :return:
        """

        return _get_message(self)


class GenericException(Arcor2Exception):
    pass


class RobotException(Arcor2Exception):
    pass


class ResourcesException(Arcor2Exception):
    pass


class ActionPointNotFound(Arcor2Exception):
    pass


class SceneObjectNotFound(Arcor2Exception):
    pass



import asyncio
from typing import AsyncIterator, Optional, Set

from arcor2 import helpers as hlp
from arcor2.cached import CachedScene, UpdateableCachedScene
from arcor2.clients import aio_persistent_storage as storage, aio_scene_service as scene_srv
from arcor2.data.common import Pose, SceneObject
from arcor2.data.object_type import Models
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Generic, GenericWithPose, Robot
from arcor2.server import globals as glob


def instances_names() -> Set[str]:
    return {obj.name for obj in glob.SCENE_OBJECT_INSTANCES.values()}


async def scenes() -> AsyncIterator[CachedScene]:

    for scene_id in (await storage.get_scenes()).items:
        yield CachedScene(await storage.get_scene(scene_id.id))


async def scene_names() -> Set[str]:
    return {scene.name for scene in (await storage.get_scenes()).items}


async def set_object_pose(obj: GenericWithPose, pose: Pose) -> None:
    """
    Object pose is property that might call scene service - that's why it should be called using executor.
    :param obj:
    :param pose:
    :return:
    """

    await hlp.run_in_executor(setattr, obj, "pose", pose)


async def add_object_to_scene(obj: SceneObject, add_to_scene: bool = True,
                              dry_run: bool = False, parent_id: Optional[str] = None) -> None:
    """

    :param obj:
    :param add_to_scene: Set to false to only create object instance and add its collision model (if any).
    :return:
    """

    assert glob.SCENE
    assert not obj.children

    if obj.type not in glob.OBJECT_TYPES:
        raise Arcor2Exception("Unknown object type.")

    obj_type = glob.OBJECT_TYPES[obj.type]

    if obj_type.meta.disabled:
        raise Arcor2Exception("Object type disabled.")

    # TODO check whether object has all required settings

    # TODO check whether object needs parent and if so, if the parent is in scene and parent_id is set
    if obj_type.meta.needs_parent_type:
        pass

    if obj_type.meta.has_pose and obj.pose is None:
        raise Arcor2Exception("Object requires pose.")

    if not obj_type.meta.has_pose and obj.pose is not None:
        raise Arcor2Exception("Object do not have pose.")

    if obj_type.meta.abstract:
        raise Arcor2Exception("Cannot instantiate abstract type.")

    if obj.id in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Object/service with that id already exists.")

    if obj.name in instances_names():
        raise Arcor2Exception("Name is already used.")

    if not hlp.is_valid_identifier(obj.name):
        raise Arcor2Exception("Object name invalid (should be snake_case).")

    if dry_run:
        return None

    glob.logger.debug(f"Creating instance {obj.id} ({obj.type}).")

    # TODO settings -> dataclass

    assert obj_type.type_def is not None

    try:

        if issubclass(obj_type.type_def, Robot):
            assert obj.pose is not None
            # TODO RPC should return here (instantiation could take long time) -> events
            glob.SCENE_OBJECT_INSTANCES[obj.id] = await hlp.run_in_executor(obj_type.type_def, obj.id, obj.name,
                                                                            obj.pose)
        elif issubclass(obj_type.type_def, GenericWithPose):
            assert obj.pose is not None
            coll_model: Optional[Models] = None
            if obj_type.meta.object_model:
                coll_model = obj_type.meta.object_model.model()

            # TODO RPC should return here (instantiation could take long time) -> events
            # TODO handle settings
            glob.SCENE_OBJECT_INSTANCES[obj.id] = await hlp.run_in_executor(
                obj_type.type_def, obj.id, obj.name, obj.pose, coll_model)

        elif issubclass(obj_type.type_def, Generic):
            assert obj.pose is None
            # TODO RPC should return here (instantiation could take long time) -> events
            # TODO handle settings
            glob.SCENE_OBJECT_INSTANCES[obj.id] = await hlp.run_in_executor(obj_type.type_def, obj.id, obj.name)

        else:
            raise Arcor2Exception("Object type with unknown base.")

    except (TypeError, ValueError) as e:  # catch some most often exceptions
        raise Arcor2Exception("Failed to create object instance.") from e

    if add_to_scene:
        glob.SCENE.upsert_object(obj)

    return None


async def clear_scene(do_cleanup: bool = True) -> None:

    glob.logger.info("Clearing the scene.")
    glob.SCENE_OBJECT_INSTANCES.clear()
    await asyncio.gather(*[hlp.run_in_executor(obj.cleanup) for obj in glob.SCENE_OBJECT_INSTANCES.values()])
    glob.SCENE = None


async def open_scene(scene_id: str) -> None:

    asyncio.ensure_future(scene_srv.delete_all_collisions())  # just for sure
    glob.SCENE = UpdateableCachedScene(await storage.get_scene(scene_id))

    try:
        await asyncio.gather(*[add_object_to_scene(obj, add_to_scene=False) for obj in glob.SCENE.objects])
    except Arcor2Exception as e:
        await clear_scene()
        raise Arcor2Exception(f"Failed to open scene. {e.message}") from e

    assert {obj.id for obj in glob.SCENE.objects} == glob.SCENE_OBJECT_INSTANCES.keys()


def get_instance(obj_id: str) -> Generic:

    if obj_id not in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Unknown object/service ID.")

    return glob.SCENE_OBJECT_INSTANCES[obj_id]



# -*- coding: utf-8 -*-

from dataclasses import dataclass, field

from arcor2.data.object_type import MeshList
from arcor2.data.rpc.common import Request, Response, wo_suffix


@dataclass
class ListMeshesRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ListMeshesResponse(Response):

    data: MeshList = field(default_factory=list)
    response: str = field(default=ListMeshesRequest.request, init=False)



import asyncio
import time
from typing import Awaitable, Callable, Dict

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2.data import common, events, rpc
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Robot
from arcor2.server import globals as glob, objects_services_actions as osa, robot
from arcor2.server.decorators import project_needed, scene_needed

TaskDict = Dict[str, asyncio.Task]

ROBOT_JOINTS_TASKS: TaskDict = {}
EEF_POSE_TASKS: TaskDict = {}

EVENT_PERIOD = 0.1


async def robot_joints_event(robot_id: str) -> None:

    global ROBOT_JOINTS_TASKS

    glob.logger.info(f"Sending '{events.RobotJointsEvent.__name__}' for robot '{robot_id}' started.")
    while glob.SCENE and glob.ROBOT_JOINTS_REGISTERED_UIS[robot_id]:

        start = time.monotonic()

        evt = events.RobotJointsEvent()

        try:
            evt.data = events.RobotJointsData(robot_id, (await robot.get_robot_joints(robot_id)))
        except Arcor2Exception as e:
            glob.logger.error(f"Failed to get joints for {robot_id}. {e.message}")
            break

        evt_json = evt.to_json()
        await asyncio.gather(*[hlp.send_json_to_client(ui, evt_json)
                               for ui in glob.ROBOT_JOINTS_REGISTERED_UIS[robot_id]])

        end = time.monotonic()
        await asyncio.sleep(EVENT_PERIOD - (end - start))

    del ROBOT_JOINTS_TASKS[robot_id]

    # TODO notify UIs that registration was cancelled
    del glob.ROBOT_JOINTS_REGISTERED_UIS[robot_id]

    glob.logger.info(f"Sending '{events.RobotJointsEvent.__name__}' for robot '{robot_id}' stopped.")


async def eef_pose(robot_id: str, eef_id: str) -> events.EefPose:

    return events.EefPose(eef_id, (await robot.get_end_effector_pose(robot_id, eef_id)))


async def robot_eef_pose_event(robot_id: str) -> None:

    global EEF_POSE_TASKS

    glob.logger.info(f"Sending '{events.RobotEefEvent.__name__}' for robot '{robot_id}' started.")

    while glob.SCENE and glob.ROBOT_EEF_REGISTERED_UIS[robot_id]:

        start = time.monotonic()

        evt = events.RobotEefEvent()
        evt.data = events.RobotEefData(robot_id)

        try:
            evt.data.end_effectors = await asyncio.gather(
                *[eef_pose(robot_id, eef_id) for eef_id in (await robot.get_end_effectors(robot_id))])
        except Arcor2Exception as e:
            glob.logger.error(f"Failed to get eef pose for {robot_id}. {e.message}")
            break

        evt_json = evt.to_json()
        await asyncio.gather(*[hlp.send_json_to_client(ui, evt_json) for ui in glob.ROBOT_EEF_REGISTERED_UIS[robot_id]])

        end = time.monotonic()
        await asyncio.sleep(EVENT_PERIOD - (end - start))

    del EEF_POSE_TASKS[robot_id]

    # TODO notify UIs that registration was cancelled
    del glob.ROBOT_EEF_REGISTERED_UIS[robot_id]

    glob.logger.info(f"Sending '{events.RobotEefEvent.__name__}' for robot '{robot_id}' stopped.")


async def get_robot_meta_cb(req: rpc.robot.GetRobotMetaRequest, ui: WsClient) ->\
        rpc.robot.GetRobotMetaResponse:

    return rpc.robot.GetRobotMetaResponse(
        data=[obj.robot_meta for obj in glob.OBJECT_TYPES.values() if obj.robot_meta is not None]
    )


@scene_needed
async def get_robot_joints_cb(req: rpc.robot.GetRobotJointsRequest, ui: WsClient) -> \
        rpc.robot.GetRobotJointsResponse:

    return rpc.robot.GetRobotJointsResponse(data=await robot.get_robot_joints(req.args.robot_id))


@scene_needed
async def get_end_effector_pose_cb(req: rpc.robot.GetEndEffectorPoseRequest, ui: WsClient) -> \
        rpc.robot.GetEndEffectorPoseResponse:

    return rpc.robot.GetEndEffectorPoseResponse(
        data=await robot.get_end_effector_pose(req.args.robot_id, req.args.end_effector_id))


@scene_needed
async def get_end_effectors_cb(req: rpc.robot.GetEndEffectorsRequest, ui: WsClient) -> \
        rpc.robot.GetEndEffectorsResponse:

    return rpc.robot.GetEndEffectorsResponse(data=await robot.get_end_effectors(req.args.robot_id))


@scene_needed
async def get_grippers_cb(req: rpc.robot.GetGrippersRequest, ui: WsClient) -> \
        rpc.robot.GetGrippersResponse:

    return rpc.robot.GetGrippersResponse(data=await robot.get_grippers(req.args.robot_id))


@scene_needed
async def get_suctions_cb(req: rpc.robot.GetSuctionsRequest, ui: WsClient) -> \
        rpc.robot.GetSuctionsResponse:

    return rpc.robot.GetSuctionsResponse(data=await robot.get_suctions(req.args.robot_id))


async def register(req: rpc.robot.RegisterForRobotEventRequest, ui: WsClient, tasks: TaskDict,
                   reg_uis: glob.RegisteredUiDict, coro: Callable[[str], Awaitable[None]]) -> None:

    if req.args.send:

        reg_uis[req.args.robot_id].add(ui)

        if req.args.robot_id not in tasks:
            # start task
            tasks[req.args.robot_id] = asyncio.create_task(coro(req.args.robot_id))

    else:
        try:
            reg_uis[req.args.robot_id].remove(ui)
        except KeyError as e:
            raise Arcor2Exception("Failed to unregister.") from e

        # cancel task if not needed anymore
        if not reg_uis[req.args.robot_id]:
            task = tasks[req.args.robot_id]

            if not task.cancelled():
                task.cancel()

            del tasks[req.args.robot_id]


@scene_needed
async def register_for_robot_event_cb(req: rpc.robot.RegisterForRobotEventRequest, ui: WsClient) -> None:

    # check if robot exists
    await osa.get_robot_instance(req.args.robot_id)

    if req.args.what == rpc.robot.RegisterEnum.JOINTS:
        await register(req, ui, ROBOT_JOINTS_TASKS, glob.ROBOT_JOINTS_REGISTERED_UIS,
                       robot_joints_event)
    elif req.args.what == rpc.robot.RegisterEnum.EEF_POSE:

        if not (await robot.get_end_effectors(req.args.robot_id)):
            raise Arcor2Exception("Robot does not have any end effector.")

        await register(req, ui, EEF_POSE_TASKS, glob.ROBOT_EEF_REGISTERED_UIS,
                       robot_eef_pose_event)
    else:
        raise Arcor2Exception(f"Option '{req.args.what.value}' not implemented.")

    return None


async def check_feature(robot_id: str, feature_name: str) -> None:

    obj_type = glob.OBJECT_TYPES[(await osa.get_robot_instance(robot_id)).__class__.__name__]

    if obj_type.robot_meta is None:
        raise Arcor2Exception("Not a robot.")

    if not getattr(obj_type.robot_meta.features, feature_name):
        raise Arcor2Exception(f"Robot does not support '{feature_name}' feature.")


@scene_needed
async def move_to_pose_cb(req: rpc.robot.MoveToPoseRequest, ui: WsClient) -> None:

    await check_feature(req.args.robot_id, Robot.move_to_pose.__name__)
    await robot.check_robot_before_move(req.args.robot_id)

    if (req.args.position is None) != (req.args.orientation is None):

        target_pose = await robot.get_end_effector_pose(req.args.robot_id, req.args.end_effector_id)

        if req.args.position:
            target_pose.position = req.args.position
        elif req.args.orientation:
            target_pose.orientation = req.args.orientation

    elif req.args.position is not None and req.args.orientation is not None:
        target_pose = common.Pose(req.args.position, req.args.orientation)
    else:
        raise Arcor2Exception("Position or orientation should be given.")

    # TODO check if the target pose is reachable (dry_run)
    asyncio.ensure_future(robot.move_to_pose(req.args.robot_id, req.args.end_effector_id, target_pose, req.args.speed))


@scene_needed
async def move_to_joints_cb(req: rpc.robot.MoveToJointsRequest, ui: WsClient) -> None:

    await check_feature(req.args.robot_id, Robot.move_to_joints.__name__)
    await robot.check_robot_before_move(req.args.robot_id)
    asyncio.ensure_future(robot.move_to_joints(req.args.robot_id, req.args.joints, req.args.speed))


@scene_needed
async def stop_robot_cb(req: rpc.robot.StopRobotRequest, ui: WsClient) -> None:

    await check_feature(req.args.robot_id, Robot.stop.__name__)
    await robot.stop(req.args.robot_id)


@scene_needed
@project_needed
async def move_to_action_point_cb(req: rpc.robot.MoveToActionPointRequest, ui: WsClient) -> None:

    await robot.check_robot_before_move(req.args.robot_id)

    assert glob.PROJECT

    # TODO check RobotMeta if the robot supports this

    if (req.args.orientation_id is None) == (req.args.joints_id is None):
        raise Arcor2Exception("Set orientation or joints. Not both.")

    if req.args.orientation_id:

        await check_feature(req.args.robot_id, Robot.move_to_pose.__name__)

        if req.args.end_effector_id is None:
            raise Arcor2Exception("eef id has to be set.")

        pose = glob.PROJECT.pose(req.args.orientation_id)

        # TODO check if the target pose is reachable (dry_run)
        asyncio.ensure_future(robot.move_to_ap_orientation(
            req.args.robot_id, req.args.end_effector_id, pose, req.args.speed, req.args.orientation_id))

    elif req.args.joints_id:

        await check_feature(req.args.robot_id, Robot.move_to_joints.__name__)

        joints = glob.PROJECT.joints(req.args.joints_id)

        # TODO check if the joints are within limits and reachable (dry_run)
        asyncio.ensure_future(robot.move_to_ap_joints(
            req.args.robot_id, joints.joints, req.args.speed, req.args.joints_id))



import asyncio
from typing import AsyncIterator, Dict, List, Set, Union

from arcor2 import helpers as hlp
from arcor2.cached import CachedProject, CachedScene, UpdateableCachedProject
from arcor2.clients import aio_persistent_storage as storage
from arcor2.data import common, events, object_type
from arcor2.exceptions import Arcor2Exception
from arcor2.parameter_plugins import PARAM_PLUGINS
from arcor2.parameter_plugins.base import ParameterPluginException
from arcor2.server import globals as glob, notifications as notif
from arcor2.server.scene import clear_scene, open_scene


async def close_project(do_cleanup: bool = True) -> None:

    glob.PROJECT = None
    await clear_scene(do_cleanup)
    asyncio.ensure_future(notif.broadcast_event(events.ProjectClosed()))


def check_action_params(scene: CachedScene, project: CachedProject, action: common.Action,
                        object_action: object_type.ObjectAction) -> None:

    _, action_type = action.parse_type()

    assert action_type == object_action.name

    if len(object_action.parameters) != len(action.parameters):
        raise Arcor2Exception("Unexpected number of parameters.")

    for req_param in object_action.parameters:

        param = action.parameter(req_param.name)

        if param.type == common.ActionParameter.TypeEnum.CONSTANT:

            const = project.constant(param.value)

            param_meta = object_action.parameter(param.id)
            if param_meta.type != const.type:
                raise Arcor2Exception("Param type does not match constant type.")

        elif param.type == common.ActionParameter.TypeEnum.LINK:

            assert param.type is None

            parsed_link = param.parse_link()
            outputs = project.action(parsed_link.action_id).flow(parsed_link.flow_name)

            assert len(outputs) == len(object_action.returns)

            param_meta = object_action.parameter(param.id)
            if param_meta.type != object_action.returns[parsed_link.output_index]:
                raise Arcor2Exception("Param type does not match action output type.")

        else:

            if param.type not in PARAM_PLUGINS:
                raise Arcor2Exception(f"Parameter {param.id} of action {action.name} has unknown type: {param.type}.")

            try:
                PARAM_PLUGINS[param.type].value(
                    {k: v.type_def for k, v in glob.OBJECT_TYPES.items() if v.type_def is not None},
                    scene, project, action.id, param.id)
            except ParameterPluginException as e:
                raise Arcor2Exception(f"Parameter {param.id} of action {action.name} has invalid value. {str(e)}")


def check_flows(parent: Union[CachedProject, common.ProjectFunction],
                action: common.Action, action_meta: object_type.ObjectAction) -> None:
    """
    Raises exception if there is something wrong with flow(s).
    :param parent:
    :param action:
    :param action_meta:
    :return:
    """

    flow = action.flow()  # searches default flow (just this flow is supported so far)

    if len(flow.outputs) != len(action_meta.returns):
        raise Arcor2Exception("Number of the flow outputs does not match the number of action outputs.")

    for output in flow.outputs:
        if not hlp.is_valid_identifier(output):
            raise Arcor2Exception(f"Output {output} is not a valid Python identifier.")

    outputs: Set[str] = set()

    for act in parent.actions:
        for fl in act.flows:
            for output in fl.outputs:
                if output in outputs:
                    raise Arcor2Exception(f"Output '{output}' is not unique.")


def find_object_action(scene: CachedScene, action: common.Action) -> object_type.ObjectAction:

    obj_id, action_type = action.parse_type()
    obj = scene.object(obj_id)

    try:
        obj_type = glob.OBJECT_TYPES[obj.type]
    except KeyError:
        raise Arcor2Exception("Unknown object type.")

    try:
        act = obj_type.actions[action_type]
    except KeyError:
        raise Arcor2Exception("Unknown type of action.")

    if act.disabled:
        raise Arcor2Exception("Action is disabled.")

    return act


async def project_names() -> Set[str]:
    return {proj.name for proj in (await storage.get_projects()).items}


async def associated_projects(scene_id: str) -> Set[str]:
    return {project.id async for project in projects(scene_id)}


async def scene_object_pose_updated(scene_id: str, obj_id: str) -> None:
    """
    Invalidates robot joints if action point's parent has changed its pose.
    :param scene_id:
    :param obj_id:
    :return:
    """

    async for project in projects_using_object_as_parent(scene_id, obj_id):

        for ap in project.action_points:

            if ap.parent != obj_id:
                continue

            glob.logger.debug(f"Invalidating joints for {project.name}/{ap.name}.")
            ap.invalidate_joints()

        await storage.update_project(project)


async def remove_object_references_from_projects(obj_id: str) -> None:

    assert glob.SCENE

    updated_project_ids: Set[str] = set()

    async for project in projects_using_object_as_parent(glob.SCENE.id, obj_id):

        # action_ids: Set[str] = set()

        for ap in project.action_points:

            # delete actions using the object
            ap.actions = [act for act in ap.actions if act.parse_type()[0] != obj_id]

            # delete actions using obj's action points as parameters
            # TODO fix this!
            """
            actions_using_invalid_param: Set[str] = \
                {act.id for act in ap.actions for param in act.parameters
                 if param.type in (ActionParameterTypeEnum.JOINTS, ActionParameterTypeEnum.POSE) and
                 param.value.startswith(obj_id)}

            ap.actions = [act for act in ap.actions if act.id not in actions_using_invalid_param]

            # get IDs of remaining actions
            action_ids.update({act.id for act in ap.actions})
            """

        # valid_ids: Set[str] = action_ids | ActionIOEnum.set()

        # TODO remove invalid logic items

        await storage.update_project(project)
        updated_project_ids.add(project.id)

    glob.logger.info("Updated projects: {}".format(updated_project_ids))


async def projects(scene_id: str) -> AsyncIterator[common.Project]:

    id_list = await storage.get_projects()

    for project_meta in id_list.items:

        project = await storage.get_project(project_meta.id)

        if project.scene_id != scene_id:
            continue

        yield project


def _project_using_object_as_parent(project: common.Project, obj_id: str) -> bool:

    for ap in project.action_points:
        if ap.parent == obj_id:
            return True

    return False


def _project_referencing_object(project: common.Project, obj_id: str) -> bool:

    for ap in project.action_points:
        for action in ap.actions:
            action_obj_id, _ = action.parse_type()

            if action_obj_id == obj_id:
                return True

    return False


async def projects_using_object(scene_id: str, obj_id: str) -> AsyncIterator[common.Project]:
    """
    Combines functionality of projects_using_object_as_parent and projects_referencing_object.
    :param scene_id:
    :param obj_id:
    :return:
    """

    async for project in projects(scene_id):
        if _project_using_object_as_parent(project, obj_id) or _project_referencing_object(project, obj_id):
            yield project


async def projects_using_object_as_parent(scene_id: str, obj_id: str) -> AsyncIterator[common.Project]:

    async for project in projects(scene_id):
        if _project_using_object_as_parent(project, obj_id):
            yield project


async def projects_referencing_object(scene_id: str, obj_id: str) -> AsyncIterator[common.Project]:

    async for project in projects(scene_id):
        if _project_referencing_object(project, obj_id):
            yield project


def project_problems(scene: CachedScene, project: CachedProject) -> List[str]:

    scene_objects: Dict[str, str] = {obj.id: obj.type for obj in scene.objects}

    action_ids: Set[str] = set()
    problems: List[str] = []

    unknown_types = {obj.type for obj in scene.objects} - glob.OBJECT_TYPES.keys()

    if unknown_types:
        return [f"Scene invalid, contains unknown types: {unknown_types}."]

    for ap in project.action_points:

        # test if all objects exists in scene
        if ap.parent and ap.parent not in scene_objects:
            problems.append(f"Action point '{ap.name}' has parent '{ap.parent}' that does not exist in the scene.")
            continue

        for joints in ap.robot_joints:
            if not joints.is_valid:
                problems.append(f"Action point {ap.name} has invalid joints: {joints.name} "
                                f"(robot {joints.robot_id}).")

        for action in ap.actions:

            if action.id in action_ids:
                problems.append(f"Action {action.name} of the {ap.name} is not unique.")

            # check if objects have used actions
            obj_id, action_type = action.parse_type()

            if obj_id not in scene_objects.keys():
                problems.append(f"Object ID {obj_id} which action is used in {action.name} does not exist in scene.")
                continue

            try:
                os_type = scene_objects[obj_id]  # object type
            except KeyError:
                os_type = obj_id  # service

            if action_type not in glob.OBJECT_TYPES[os_type].actions:
                problems.append(f"Object type {scene_objects[obj_id]} does not have action {action_type} "
                                f"used in {action.id}.")
                continue

            try:
                check_action_params(scene, project, action, glob.OBJECT_TYPES[os_type].actions[action_type])
            except Arcor2Exception as e:
                problems.append(e.message)

    return problems


async def open_project(project_id: str) -> None:

    project = UpdateableCachedProject(await storage.get_project(project_id))

    if glob.SCENE:
        if glob.SCENE.id != project.scene_id:
            raise Arcor2Exception("Required project is associated to another scene.")
    else:
        await open_scene(project.scene_id)

    assert glob.SCENE
    for ap in project.action_points_with_parent:

        assert ap.parent
        try:
            glob.SCENE.object(ap.parent)
        except Arcor2Exception:
            await clear_scene()
            raise Arcor2Exception(f"Action point's {ap.name} parent not available in the scene.")

    glob.PROJECT = project



# -*- coding: utf-8 -*-

from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data.common import ActionParameter, Flow, IdDesc, Orientation, Position, ProjectLogicIf
from arcor2.data.rpc.common import IdArgs, Request, Response, RobotArg, wo_suffix


@dataclass
class NewProjectRequestArgs(JsonSchemaMixin):

    scene_id: str
    name: str
    desc: str = field(default_factory=str)
    has_logic: bool = True


@dataclass
class NewProjectRequest(Request):

    args: NewProjectRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class NewProjectResponse(Response):

    response: str = field(default=NewProjectRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class CloseProjectRequestArgs(JsonSchemaMixin):

    force: bool = False


@dataclass
class CloseProjectRequest(Request):

    args: CloseProjectRequestArgs = field(default_factory=CloseProjectRequestArgs)
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class CloseProjectResponse(Response):

    response: str = field(default=CloseProjectRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class SaveProjectRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class SaveProjectResponse(Response):

    response: str = field(default=SaveProjectRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class OpenProjectRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class OpenProjectResponse(Response):

    response: str = field(default=OpenProjectRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ListProjectsResponseData(IdDesc):

    scene_id: str = field(default_factory=str)
    valid: bool = field(default=False, metadata=dict(description="Objects and their actions exists."))
    executable: bool = field(default=False, metadata=dict(description="Logic is defined and valid."))
    problems: List[str] = field(default_factory=list)
    modified: Optional[datetime] = None


@dataclass
class ListProjectsRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ListProjectsResponse(Response):

    data: List[ListProjectsResponseData] = field(default_factory=list)
    response: str = field(default=ListProjectsRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class ExecuteActionArgs(JsonSchemaMixin):

    action_id: str = field(metadata=dict(description="ID of the action to be executed."))


@dataclass
class ExecuteActionRequest(Request):

    args: ExecuteActionArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class ExecuteActionResponse(Response):

    response: str = field(default=ExecuteActionRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------

@dataclass
class CancelActionRequest(Request):

    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class CancelActionResponse(Response):

    response: str = field(default=CancelActionRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class AddActionPointArgs(JsonSchemaMixin):

    name: str
    position: Position
    parent: Optional[str] = None


@dataclass
class AddActionPointRequest(Request):

    args: AddActionPointArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AddActionPointResponse(Response):

    response: str = field(default=AddActionPointRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RemoveActionPointRequest(Request):

    args: IdArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RemoveActionPointResponse(Response):

    response: str = field(default=RemoveActionPointRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class AddActionPointJointsRequestArgs(JsonSchemaMixin):

    action_point_id: str
    robot_id: str
    name: str = "default"


@dataclass
class AddActionPointJointsRequest(Request):

    args: AddActionPointJointsRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AddActionPointJointsResponse(Response):

    response: str = field(default=AddActionPointJointsRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RenameActionPointRequestArgs(JsonSchemaMixin):

    action_point_id: str
    new_name: str


@dataclass
class RenameActionPointRequest(Request):

    args: RenameActionPointRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RenameActionPointResponse(Response):

    response: str = field(default=RenameActionPointRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateActionPointParentRequestArgs(JsonSchemaMixin):

    action_point_id: str
    new_parent_id: str


@dataclass
class UpdateActionPointParentRequest(Request):

    args: UpdateActionPointParentRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateActionPointParentResponse(Response):

    response: str = field(default=UpdateActionPointParentRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateActionPointPositionRequestArgs(JsonSchemaMixin):

    action_point_id: str
    new_position: Position


@dataclass
class UpdateActionPointPositionRequest(Request):

    args: UpdateActionPointPositionRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateActionPointPositionResponse(Response):

    response: str = field(default=UpdateActionPointPositionRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateActionPointUsingRobotRequestArgs(JsonSchemaMixin):

    action_point_id: str
    robot: RobotArg


@dataclass
class UpdateActionPointUsingRobotRequest(Request):

    args: UpdateActionPointUsingRobotRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateActionPointUsingRobotResponse(Response):

    response: str = field(default=UpdateActionPointUsingRobotRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class AddActionPointOrientationUsingRobotRequestArgs(JsonSchemaMixin):

    action_point_id: str
    robot: RobotArg
    name: str = "default"


@dataclass
class AddActionPointOrientationUsingRobotRequest(Request):

    args: AddActionPointOrientationUsingRobotRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AddActionPointOrientationUsingRobotResponse(Response):

    response: str = field(default=AddActionPointOrientationUsingRobotRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateActionPointOrientationUsingRobotRequestArgs(JsonSchemaMixin):

    action_point_id: str
    robot: RobotArg
    orientation_id: str


@dataclass
class UpdateActionPointOrientationUsingRobotRequest(Request):

    args: UpdateActionPointOrientationUsingRobotRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateActionPointOrientationUsingRobotResponse(Response):

    response: str = field(default=UpdateActionPointOrientationUsingRobotRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class AddActionPointOrientationRequestArgs(JsonSchemaMixin):

    action_point_id: str
    orientation: Orientation
    name: str = "default"


@dataclass
class AddActionPointOrientationRequest(Request):

    args: AddActionPointOrientationRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AddActionPointOrientationResponse(Response):

    response: str = field(default=AddActionPointOrientationRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateActionPointOrientationRequestArgs(JsonSchemaMixin):

    orientation_id: str
    orientation: Orientation


@dataclass
class UpdateActionPointOrientationRequest(Request):

    args: UpdateActionPointOrientationRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateActionPointOrientationResponse(Response):

    response: str = field(default=UpdateActionPointOrientationRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RemoveActionPointOrientationRequestArgs(JsonSchemaMixin):
    action_point_id: str
    orientation_id: str


@dataclass
class RemoveActionPointOrientationRequest(Request):
    args: RemoveActionPointOrientationRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RemoveActionPointOrientationResponse(Response):
    response: str = field(default=RemoveActionPointOrientationRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class AddActionRequestArgs(JsonSchemaMixin):

    action_point_id: str
    name: str
    type: str
    parameters: List[ActionParameter] = field(default_factory=list)
    flows: List[Flow] = field(default_factory=list)


@dataclass
class AddActionRequest(Request):

    args: AddActionRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AddActionResponse(Response):

    response: str = field(default=AddActionRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateActionRequestArgs(JsonSchemaMixin):

    action_id: str
    parameters: Optional[List[ActionParameter]] = None
    flows: Optional[List[Flow]] = None


@dataclass
class UpdateActionRequest(Request):

    args: UpdateActionRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateActionResponse(Response):

    response: str = field(default=UpdateActionRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RemoveActionRequest(Request):

    args: IdArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RemoveActionResponse(Response):

    response: str = field(default=RemoveActionRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class AddLogicItemArgs(JsonSchemaMixin):

    start: str
    end: str
    condition: Optional[ProjectLogicIf] = None


@dataclass
class AddLogicItemRequest(Request):

    args: AddLogicItemArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AddLogicItemResponse(Response):

    response: str = field(default=AddLogicItemRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateLogicItemArgs(JsonSchemaMixin):

    logic_item_id: str
    start: str
    end: str
    condition: Optional[ProjectLogicIf] = None


@dataclass
class UpdateLogicItemRequest(Request):

    args: UpdateLogicItemArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateLogicItemResponse(Response):

    response: str = field(default=UpdateLogicItemRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RemoveLogicItemArgs(JsonSchemaMixin):

    logic_item_id: str


@dataclass
class RemoveLogicItemRequest(Request):

    args: RemoveLogicItemArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RemoveLogicItemResponse(Response):

    response: str = field(default=RemoveLogicItemRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateActionPointJointsRequestArgs(JsonSchemaMixin):

    robot_id: str
    joints_id: str = "default"


@dataclass
class UpdateActionPointJointsRequest(Request):

    args: UpdateActionPointJointsRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateActionPointJointsResponse(Response):

    response: str = field(default=UpdateActionPointJointsRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RemoveActionPointJointsRequestArgs(JsonSchemaMixin):

    joints_id: str


@dataclass
class RemoveActionPointJointsRequest(Request):

    args: RemoveActionPointJointsRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RemoveActionPointJointsResponse(Response):

    response: str = field(default=RemoveActionPointJointsRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class DeleteProjectRequest(Request):

    args: IdArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class DeleteProjectResponse(Response):

    response: str = field(default=DeleteProjectRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RenameProjectRequestArgs(JsonSchemaMixin):

    project_id: str
    new_name: str


@dataclass
class RenameProjectRequest(Request):

    args: RenameProjectRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RenameProjectResponse(Response):

    response: str = field(default=RenameProjectRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class CopyProjectRequestArgs(JsonSchemaMixin):

    source_id: str
    target_name: str


@dataclass
class CopyProjectRequest(Request):

    args: CopyProjectRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class CopyProjectResponse(Response):

    response: str = field(default=CopyProjectRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateProjectDescriptionRequestArgs(JsonSchemaMixin):

    project_id: str
    new_description: str


@dataclass
class UpdateProjectDescriptionRequest(Request):

    args: UpdateProjectDescriptionRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateProjectDescriptionResponse(Response):

    response: str = field(default=UpdateProjectDescriptionRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateProjectHasLogicRequestArgs(JsonSchemaMixin):

    project_id: str
    new_has_logic: bool


@dataclass
class UpdateProjectHasLogicRequest(Request):

    args: UpdateProjectHasLogicRequestArgs
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateProjectHasLogicResponse(Response):

    response: str = field(default=UpdateProjectHasLogicRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RenameActionPointJointsRequestArgs(JsonSchemaMixin):

    joints_id: str
    new_name: str


@dataclass
class RenameActionPointJointsRequest(Request):

    args: RenameActionPointJointsRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RenameActionPointJointsResponse(Response):

    response: str = field(default=RenameActionPointJointsRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RenameActionPointOrientationRequestArgs(JsonSchemaMixin):

    orientation_id: str
    new_name: str


@dataclass
class RenameActionPointOrientationRequest(Request):

    args: RenameActionPointOrientationRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RenameActionPointOrientationResponse(Response):

    response: str = field(default=RenameActionPointOrientationRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RenameActionRequestArgs(JsonSchemaMixin):

    action_id: str
    new_name: str


@dataclass
class RenameActionRequest(Request):

    args: RenameActionRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RenameActionResponse(Response):

    response: str = field(default=RenameActionRequest.request, init=False)


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class AddConstantRequestArgs(JsonSchemaMixin):

    name: str
    type: str
    value: str


@dataclass
class AddConstantRequest(Request):

    args: AddConstantRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class AddConstantResponse(Response):

    response: str = field(default=AddConstantRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class UpdateConstantRequestArgs(JsonSchemaMixin):

    constant_id: str
    name: Optional[str] = None
    value: Optional[str] = None


@dataclass
class UpdateConstantRequest(Request):

    args: UpdateConstantRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class UpdateConstantResponse(Response):

    response: str = field(default=UpdateConstantRequest.request, init=False)

# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class RemoveConstantRequestArgs(JsonSchemaMixin):

    constant_id: str


@dataclass
class RemoveConstantRequest(Request):

    args: RemoveConstantRequestArgs
    dry_run: bool = False
    request: str = field(default=wo_suffix(__qualname__), init=False)  # type: ignore  # noqa: F821


@dataclass
class RemoveConstantResponse(Response):

    response: str = field(default=RemoveConstantRequest.request, init=False)



from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data import common, object_type


@dataclass
class PackageInfo(JsonSchemaMixin):

    package_id: str
    package_name: str
    scene: common.Scene
    project: common.Project
    collision_models: object_type.CollisionModels = field(default_factory=object_type.CollisionModels)


@dataclass
class PackageMeta(JsonSchemaMixin):

    name: str
    built: datetime
    executed: Optional[datetime] = None


@dataclass
class PackageSummary(JsonSchemaMixin):

    id: str
    project_id: str
    modified: datetime = field(metadata=dict(
        description="Last modification of the project embedded in the execution package."))
    package_meta: PackageMeta = field(metadata=dict(description="Content of 'package.json'."))



import asyncio

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.data import rpc
from arcor2.exceptions import Arcor2Exception
from arcor2.server import decorators, globals as glob
from arcor2.server.execution import build_and_upload_package, run_temp_package


async def build_project_cb(req: rpc.execution.BuildProjectRequest, ui: WsClient) -> rpc.execution.BuildProjectResponse:
    """
    Builds project and uploads resulting package to the execution unit.
    :param req:
    :param ui:
    :return:
    """

    package_id = await build_and_upload_package(req.args.project_id, req.args.package_name)

    resp = rpc.execution.BuildProjectResponse()
    resp.data = rpc.execution.BuildProjectData(package_id)
    return resp


@decorators.project_needed
async def temporary_package_cb(req: rpc.execution.TemporaryPackageRequest, ui: WsClient) -> None:

    assert glob.PROJECT

    if glob.PROJECT.has_changes:
        raise Arcor2Exception("Project has unsaved changes.")

    package_id = await build_and_upload_package(glob.PROJECT.id,
                                                f"Temporary package for project '{glob.PROJECT.name}'.")

    asyncio.ensure_future(run_temp_package(package_id))
    return None
